/*

  Doom 64 Macro Compiler

  Copyright 2008 James Haley

  haleyjd@hotmail.com

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include <cstring>

#include "utils.h"
#include "macrodata.h"

#define NUMLINESPECS 256

// Note: All specials named with a number are unknown. A bounty is offered
// for any info pertaining to said specials. The bounty is my eternal 
// gratitude ;)

const char *specialDef[NUMLINESPECS] =
{
   "None",
   "DoorOpenClose",
   "DoorOpen",
   "DoorClose",
   "DoorOpenClassic",
   "FloorRaise",
   "CeilingCrushRaiseFast",
   "7",
   "StairsRaise8",
   "9",
   "PlatformDownUp",
   "11",
   "12",
   "13",
   "14",
   "15",
   "16",
   "17",
   "FloorRaiseNearest",
   "FloorLower",
   "20",
   "PlatformDownUp2", // FIXME
   "22",
   "FloorLowerLowest",
   "24",
   "CeilingCrushRaise",
   "26",
   "27",
   "28",
   "29",
   "30",
   "DoorOpenCloseClassic",
   "32",
   "33",
   "34",
   "35",
   "FloorLowerFast",
   "FloorLowerChange",
   "FloorLowerLowest2", // FIXME
   "Teleport",
   "40",
   "41",
   "42",
   "43",
   "44",
   "45",
   "46",
   "47",
   "48",
   "49",
   "50",
   "51",
   "ExitLevel",
   "PlatformPerpetualRaise",
   "PlatformPerpetualStop",
   "55",
   "56",
   "CeilingCrushStop",
   "58",
   "59",
   "60",
   "61",
   "62",
   "63",
   "64",
   "65",
   "66",
   "67",
   "68",
   "69",
   "70",
   "71",
   "72",
   "73",
   "74",
   "75",
   "76",
   "77",
   "78",
   "79",
   "80",
   "81",
   "82",
   "83",
   "84",
   "85",
   "86",
   "87",
   "88",
   "89",
   "ArtifactSwitch1",
   "ArtifactSwitch2",
   "ArtifactSwitch3",
   "93",
   "AlertEnemies",
   "95",
   "96",
   "97",
   "98",
   "99",
   "StairsRaiseFast16",
   "101",
   "102",
   "103",
   "104",
   "105",
   "106",
   "107",
   "DoorOpenFast",
   "DoorOpenFast2", // FIXME
   "DoorCloseFast",
   "111",
   "112",
   "113",
   "114",
   "115",
   "116",
   "DoorOpenCloseFast",
   "DoorOpenCloseFastClassic",
   "119",
   "120",
   "PlatformDownUpFast",
   "PlatformUpDown",
   "PlatformUpDownFast",
   "SecretExit",
   "TeleportNonPlayer",
   "126",
   "127",
   "128",
   "129",
   "130",
   "131",
   "132",
   "133",
   "134",
   "135",
   "136",
   "137",
   "138",
   "139",
   "140",
   "KillMonstersInSectorSpecial666",
   "142",
   "143",
   "144",
   "145",
   "146",
   "147",
   "148",
   "149",
   "150",
   "151",
   "152",
   "153",
   "154",
   "155",
   "156",
   "157",
   "158",
   "159",
   "160",
   "161",
   "162",
   "163",
   "164",
   "165",
   "166",
   "167",
   "168",
   "169",
   "170",
   "171",
   "172",
   "173",
   "174",
   "175",
   "176",
   "177",
   "178",
   "179",
   "180",
   "181",
   "182",
   "183",
   "184",
   "185",
   "186",
   "187",
   "188",
   "189",
   "190",
   "191",
   "192",
   "CustomFloorMove",
   "194",
   "195",
   "196",
   "197",
   "198",
   "199",
   "ClearCamera",
   "SetCamera",
   "InvokeDart",
   "Delay",
   "SetLevelInteger", //Copies a value to a global integer.
   "205",
   "206",
   "207",
   "208",
   "209",
   "CustomCeiling",
   "211",
   "CustomFloor",
   "213",
   "FloorCeilingMove",
   "215",
   "216",
   "217",
   "SetLineFlag",
   "SetLineTexture",
   "SetSectorFlag",
   "SetSectorSpecial",
   "SetSectorLight",
   "SetSectorFlats",
   "SpawnThing",
   "Quake",
   "CustomCeilingFast",
   "SetCeilingHeight",
   "CustomFloorFast",
   "SetFloorHeight",
   "SetLineSpecial",
   "InvokeTracer",
   "CeilingCrushRaiseOnce",
   "FreezePlayer",
   "234",
   "SetLightData",
   "CustomPlatformDownUp",
   "CustomPlatformDownUpFast",
   "CustomPlatformUpDown",
   "CustomPlatformUpDownFast",
   "TriggerRandomTaggedLine",
   "CeilingFloorSplit",
   "RemoveThing",
   "MoveCamera",
   "SetSectorFloorExact",
   "SetSectorCeilingExact",
   "RestartMacro",
   "CustomFloorToHeight",
   "248",
   "Telefrag",
   "250",
   "251",
   "CustomCeilingToHeight",
   "EnableFeatures",
   "Map33Special",
   "255"
};

// this and the above could be combined, but I am lazy ;)
// it also keeps the names marginally easier to edit in the future.

struct LineSpecData
{
   const char *name;
   int hashfirst;
   int hashnext;
};

static LineSpecData linespecs[NUMLINESPECS+1]; // how nice that 257 is prime ;)

//
// InitLineSpecData
//
// Sets up a threaded hash table for the special names.
//
static void InitLineSpecData(void)
{
   int i;

   // init all hashnext's to -1
   for(i = 0; i < NUMLINESPECS + 1; ++i)
   {
      linespecs[i].hashfirst = -1;
      linespecs[i].hashnext  = -1;
   }

   for(i = 0; i < NUMLINESPECS; ++i)
   {
      unsigned int key = StringHash(specialDef[i]) % (NUMLINESPECS + 1);

      linespecs[i].name = specialDef[i];
      
      // link into thread at linespecs[key]
      linespecs[i].hashnext = linespecs[key].hashfirst;
      linespecs[key].hashfirst = i;
   }
}

//
// D64_LineSpecNumForName
//
// Given a name, returns the corresponding special number.
// If the name is not found, 0 is returned.
//
int D64_LineSpecNumForName(const char *name)
{
   static bool firsttime = true;
   int i;

   if(firsttime)
   {
      firsttime = false;
      InitLineSpecData();
   }

   i = linespecs[StringHash(name) % (NUMLINESPECS+1)].hashfirst;

   while(i >= 0 && strcasecmp(name, linespecs[i].name))
      i = linespecs[i].hashnext;

   return i;
}

//
// D64_LineSpecNameForNum
//
// Given a special number between 0 and 255, returns the corresponding name.
// Returns NULL otherwise.
//
const char *D64_LineSpecNameForNum(int i)
{
   return (i >= 0 && i < NUMLINESPECS) ? specialDef[i] : NULL;
}

// EOF

